#!/bin/sh -e

if [ -f .git/hooks/pre-commit.sample -a ! -f .git/hooks/pre-commit ] ; then
        cp -p .git/hooks/pre-commit.sample .git/hooks/pre-commit && \
        chmod +x .git/hooks/pre-commit && \
        echo "Activated pre-commit hook."
fi

autoreconf --install --symlink

echo
echo "--------------------------------------------------------"
echo "Initialized build system.  For a default build, execute:"
echo "--------------------------------------------------------"
echo
echo "./configure && make"
echo
